from os.path import commonprefix


__all__ = (
    "DataSource",
    "Preprocessor",
    "LearnblockIdentifier"
)


class DataSource:
    def __init__(self, source, learnblock_identifier, *, block_size):
        self.block_size = block_size
        self.__source = source
        self.__learnblock_identifier = learnblock_identifier

    @property
    def learnblocks(self):
        for block in self:
            for learnblock in self.__learnblock_identifier.identify(block):
                yield learnblock

    def __getitem__(self, item):
        start_index = item*self.block_size
        stop_index = (item+1)*self.block_size

        if stop_index > self.__source.length:
            stop_index = None

        if start_index > self.__source.length:
            raise IndexError()

        return self.__source.get_block(start_index, stop_index)

    def __len__(self):
        return self.__source.length


class Preprocessor:
    TIME_COLUMN = "T"
    TARGET_COLUMN = "Z"
    SIGMA_COLUMN = "Sigma"

    def __init__(self, settings):
        self.settings = settings

    def clean(self, table):
        self._drop_irrelevant_columns(table)

        if self.settings.set_targets:
            self._overwrite_target_column(table)

        if self.settings.sort_time_stamp:
            self._sort_according_time_stamp(table)

        if self.settings.cut_time_stamp:
            self._remove_common_time_stamp_prefix(table)

        return table

    def _drop_irrelevant_columns(self, table):
        features_to_be_removed = [table.get_column_name_by_index(i)
                                  for i in self.settings.set_features]
        for column in features_to_be_removed:
            table.drop_column_by_name(column)

    def _overwrite_target_column(self, table):
        table.set_column_value(self.TARGET_COLUMN, self.settings.set_targets)

    def _sort_according_time_stamp(self, table):
        table.sort(self.TIME_COLUMN)

    def _remove_common_time_stamp_prefix(self, table):
        # TODO (dmt): Check if timestamp column is of type string!
        time_column = table.get_column_values_as_list(self.TIME_COLUMN)
        common_prefix = commonprefix(time_column)
        cleaned_time_column = [s.lstrip(common_prefix) for s in time_column]
        table.set_column_values(cleaned_time_column)


class LearnblockIdentifier:
    def __init__(self, settings):
        self.settings = settings
        self.column_pairs = (("T", "Z"), ("T", "Sigma"), ("Sigma", "Z"))

    def identify(self, block):
        for pair in self.column_pairs:
            for possible_learnblock in self._identify_relatives(block, *pair):
                if self._is_learn_block(possible_learnblock.length):
                    yield possible_learnblock

    def _is_learn_block(self, block_length):
        return block_length > self.settings.learn_block_minimum

    def _identify_relatives(self, block, *args):
        # TODO (dmt): Implement density and kmeans!
        already_seen = set()
        for value_pair in block.get_duplicated_pairs(args[0], args[1]):
            if value_pair not in already_seen:
                already_seen.add(value_pair)
                kw = {args[0]: value_pair[0], args[1]: value_pair[1]}
                yield block.get_values(**kw)
